
extern "C"
{
#include <audacious/util.h>
#include <audacious/configdb.h>
#include <audacious/plugin.h>
#include <audacious/output.h>
#include <audacious/i18n.h>
}
#include <unistd.h>
#include <pthread.h>
#include "include/win_wrap.h"
#include "include/gui.h"
#include "include/decode.h"
#include "include/messagequeue.h"
#include "include/settings.h"
#include "include/version.h"
#include "include/data.h"

#define TM_QUIT 0
#define TM_PLAY 1
#define TM_SEEK 2

static CDecoder decoder;
static volatile long decode_seek;
static GThread *decode_thread;

static InputPlugin *cube_iplist[] = { &cube_iplug, NULL };

DECLARE_PLUGIN(cube,NULL,NULL,cube_iplist,NULL,NULL,NULL,NULL,NULL);

#define DS_EXIT -2

extern "C" void* cube_play_loop(InputPlayback *playback)
{
  unsigned char buffer[576*4];
  long l;
  
  decode_seek = -1;
  playback->playing = 1;
  playback->eof = 0;
  
  while (playback->playing)
  {
    // ******************************************
    // Seeking
    // ******************************************
    // check thread flags, not my favorite method
    if (decode_seek == DS_EXIT)
    {
      goto exit_thread;
    }
    else if (decode_seek >= 0)
    {
      if (decoder.Seek(decode_seek))
      {
	playback->output->flush(decode_seek);
	// reset eof flag
	playback->eof = 0;
      }
      else
      {
	// couldn't seek, so eof
	playback->eof = 1;
      }
      // reset decode_seek
      decode_seek = -1;
    }
    
    // ******************************************
    // Playback
    // ******************************************
    if (!playback->eof)
    {
      // read data and pass onward
      l = decoder.Get576Samples(reinterpret_cast<short*>(buffer));
      if (!l)
      {
	playback->eof = 1;
	// will trigger on next run through
      }
      else
      {
	// ok we read stuff , pass it on
	playback->pass_audio(playback,(decoder.GetBitsPerSample() == 16) ? FMT_S16_LE : FMT_U8,decoder.GetChannels() , l , buffer , &playback->playing );
      }
    }
    else
    {
      // at EOF
      playback->output->buffer_free();
      playback->output->buffer_free();
      while (playback->output->buffer_playing())
	g_usleep(10000);
      playback->playing = 0;
      // this effectively ends the loop
    }
  }
 exit_thread:
  decode_seek = -1;
  playback->playing = 0;
  decoder.Destroy();
  return 0;
}

extern "C" void cube_about()
{
  cube_gui_about();
}

extern "C" void cube_configure()
{
  cube_gui_configure();
}

extern "C" void cube_init()
{
  SETTINGS s;
  // should never fail now
  if (LoadSettings(&s))
  {
    looptimes  = s.looptimes;
    fadelength = s.fadelength;
    fadedelay  = s.fadedelay;
    adx_cutoff = s.ADXCutoff;
    adxonechan = s.ADXChannel;
    RSFCOEF1 =   s.rsf_coefficient[0];
    RSFCOEF2 =   s.rsf_coefficient[1];
  }

  decoder.Init();
}

extern "C" void cube_destroy()
{
  decoder.Destroy();
}

extern "C" gboolean cube_is_our_file(char *pFile)
{
  if (!pFile)
    return FALSE;

  return (CDecoder::CanPlay(pFile) ? TRUE : FALSE);
}

extern "C" void cube_play(InputPlayback *context)
{
  // this is now called in a new thread context

  char title[256];
  
  if (!decoder.Open(context->filename))
  {
    DisplayError("Failed to open %s",context->filename);
    return;
  }
  // open the audio device
  if (context->output->open_audio((decoder.GetBitsPerSample() == 16) ? FMT_S16_LE : FMT_U8,decoder.GetFrequency(),decoder.GetChannels()) == 0)
  {
    decoder.Destroy();
    DisplayError("Failed to open audio output (check your output plugin configuration)");
    return;
  }
  // set the info
  if (!decoder.GetLoadedFileTitle(title))
    strcpy(title,context->filename);
  
  context->set_params(context,title,
		      /* length */ decoder.GetLength(),
		      /* rate */decoder.GetBitsPerSecond(),
		      /* freq */decoder.GetFrequency(),
		      /* n channels */decoder.GetChannels());
  
  decode_thread = g_thread_self();
  context->set_pb_ready(context);
  cube_play_loop(context);
}

extern "C" void cube_stop(InputPlayback *context)
{
  if (decoder.IsLoaded())
  {
    // kill thread
    decode_seek = DS_EXIT;
    // wait for it to die
    g_thread_join(decode_thread);
    // close audio output
  }
  context->output->close_audio();
  // cleanup 
  decoder.Destroy();
}

extern "C" void cube_seek(InputPlayback *context, int time)
{
  cube_mseek(context,time * 1000);
}

extern "C" void cube_pause(InputPlayback *context, short paused)
{
  context->output->pause(paused);
}

extern "C" int cube_get_time(InputPlayback *context)
{
  if (!decoder.IsLoaded())
    return -2;
  
  if (!context->playing || 
      (decoder.IsEOF() && !context->output->buffer_playing()))
    return -1;
  
  return context->output->output_time();
}

extern "C" void cube_get_song_info(char *pFile,char **title,int *length)
{
  char *name = NULL;
  CDecoder decoder;
  if (decoder.Open(pFile))
  {
    if ((name = reinterpret_cast<char*>(g_malloc(256))))
    {
      decoder.GetLoadedFileTitle(name);
    }
    *length = decoder.GetLength();
  }
  else
  {
    *length = 0;
  }
  
  *title = name;
}

extern "C" void cube_file_info_box(char *pFile)
{
  char msg[512];
  CDecoder decoder;
  if (decoder.Open(pFile))
  {
    sprintf(msg,"%s\nSample rate: %d\nStereo: %s\nTotal samples: %d",pFile,
	    decoder.GetFrequency(),(decoder.GetChannels() == 2) ? "yes" : "no",
	    decoder.GetNumberOfSamples());
    

    audacious_info_dialog("File information",msg,"OK",FALSE,NULL,NULL);
  }
}

extern "C" void cube_mseek(InputPlayback *data,gulong ms)
{
  if (decoder.IsLoaded())
  {
    decode_seek = ms;
    data->eof = 0;

    while (decode_seek != -1)
      g_usleep(10000);
  }
}
